/*
 config.c - Config for Shutter Bridge

 Copyright (c) 2022 Sasapea's Lab. All right reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General
 Public License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <syslog.h>
#include "config.h"

static const char DEVICE_ADDRESS[]         = "DEVICE.ADDRESS";
static const char DEVICE_PULSE_TIME[]      = "DEVICE.PULSE_TIME";
static const char DEVICE_PULSE_INTERVAL[]  = "DEVICE.PULSE_INTERVAL";
static const char DEVICE_OPEN_INIT[]       = "DEVICE.OPEN_INIT";
static const char DEVICE_OPEN_TIME[]       = "DEVICE.OPEN_TIME";
static const char DEVICE_RESUME_TIME[]     = "DEVICE.RESUME_TIME";
static const char DEVICE_AUTODW_TIME[]     = "DEVICE.AUTODW_TIME";
static const char DEVICE_ALARM_DISABLE[]   = "DEVICE.ALARM_DISABLE";

static const char DEFAULT_PULSE_TIME[]     = "DEFAULT.PULSE_TIME";
static const char DEFAULT_PULSE_INTERVAL[] = "DEFAULT.PULSE_INTERVAL";
static const char DEFAULT_OPEN_INIT[]      = "DEFAULT.OPEN_INIT";
static const char DEFAULT_OPEN_TIME[]      = "DEFAULT.OPEN_TIME";
static const char DEFAULT_RESUME_TIME[]    = "DEFAULT.RESUME_TIME";
static const char DEFAULT_AUTODW_TIME[]    = "DEFAULT.AUTODW_TIME";
static const char DEFAULT_ALARM_DISABLE[]  = "DEFAULT.ALARM_DISABLE";

device_t devices[MAX_SUPPORT_DEVICES];

static setting_t defaults =
{
  .pulse_time     = 1000,  // pulse time
  .pulse_interval = 100,   // pulse interval
  .open_init      = 0,     // open init
  .open_time      = 28000, // open time
  .resume_time    = 15000, // resume time
  .autodw_time    = 30000, // autodw time
  .alarm_disable  = 0      // alarm disable
};

static void macaddr(macaddr_t& mac, const char *val)
{
  char hex[3];
  for (int i = 0; i < sizeof(mac); ++i)
  {
    hex[0] = *val++;
    hex[1] = *val++;
    hex[2] = 0;
    mac.data[i] = strtol(hex, NULL, 16);
    if (*val == ':')
      val++;
  }
}

void load_config(const char *path)
{
  FILE *fp;
  char *p, *q, *r, *e, buf[1024];
  int num, i;

  for (i = 0; i < MAX_SUPPORT_DEVICES; ++i)
    memset(&devices[i].setting, 0xFF, sizeof(devices[i].setting));

  snprintf(buf, sizeof(buf), CONFIGURATION_FILE, path);
  if ((fp = fopen(buf, "rt")) != NULL)
  {
    memset(buf, 0, sizeof(buf));
    while (fgets(buf, sizeof(buf) - 1, fp))
    {
      p = buf;
      while (*p == ' ')
        ++p;
      if (isalpha(*p))
      {
        q = p + 1;
        r = q;
        while (*q && (*q != '='))
        {
          if (*q > ' ')
            r = ++q;
          else
            ++q;
        }
        *r = 0;
        *q++ = 0;
        while (*q && (*q <= ' '))
          ++q;
        r = q;
        while (*r && (*r > ' '))
          ++r;
        *r = 0;
        //
        // device config
        //
        if (strncmp(p, DEVICE_ADDRESS, sizeof(DEVICE_ADDRESS) - 1) == 0)
        {
          num = strtol(p + sizeof(DEVICE_ADDRESS) - 1, &e, 10);
          if (!*e && (num > 0) && (num <= MAX_SUPPORT_DEVICES))
            macaddr(devices[num - 1].mac, q);
        }
        else if (strncmp(p, DEVICE_PULSE_TIME    , sizeof(DEVICE_PULSE_TIME    ) - 1) == 0)
        {
          num = strtol(p + sizeof(DEVICE_PULSE_TIME) - 1, &e, 10);
          if (!*e && (num > 0) && (num <= MAX_SUPPORT_DEVICES))
            devices[num - 1].setting.pulse_time = atoi(q);
        }
        else if (strncmp(p, DEVICE_PULSE_INTERVAL, sizeof(DEVICE_PULSE_INTERVAL) - 1) == 0)
        {
          num = strtol(p + sizeof(DEVICE_PULSE_INTERVAL) - 1, &e, 10);
          if (!*e && (num > 0) && (num <= MAX_SUPPORT_DEVICES))
            devices[num - 1].setting.pulse_interval = atoi(q);
        }
        else if (strncmp(p, DEVICE_OPEN_INIT     , sizeof(DEVICE_OPEN_INIT     ) - 1) == 0)
        {
          num = strtol(p + sizeof(DEVICE_OPEN_INIT) - 1, &e, 10);
          if (!*e && (num > 0) && (num <= MAX_SUPPORT_DEVICES))
            devices[num - 1].setting.open_init = atoi(q);
        }
        else if (strncmp(p, DEVICE_OPEN_TIME     , sizeof(DEVICE_OPEN_TIME     ) - 1) == 0)
        {
          num = strtol(p + sizeof(DEVICE_OPEN_TIME) - 1, &e, 10);
          if (!*e && (num > 0) && (num <= MAX_SUPPORT_DEVICES))
            devices[num - 1].setting.open_time = atoi(q);
        }
        else if (strncmp(p, DEVICE_RESUME_TIME   , sizeof(DEVICE_RESUME_TIME   ) - 1) == 0)
        {
          num = strtol(p + sizeof(DEVICE_RESUME_TIME) - 1, &e, 10);
          if (!*e && (num > 0) && (num <= MAX_SUPPORT_DEVICES))
            devices[num - 1].setting.resume_time = atoi(q);
        }
        else if (strncmp(p, DEVICE_AUTODW_TIME   , sizeof(DEVICE_AUTODW_TIME   ) - 1) == 0)
        {
          num = strtol(p + sizeof(DEVICE_AUTODW_TIME) - 1, &e, 10);
          if (!*e && (num > 0) && (num <= MAX_SUPPORT_DEVICES))
            devices[num - 1].setting.autodw_time = atoi(q);
        }
        else if (strncmp(p, DEVICE_ALARM_DISABLE, sizeof(DEVICE_ALARM_DISABLE  ) - 1) == 0)
        {
          num = strtol(p + sizeof(DEVICE_ALARM_DISABLE) - 1, &e, 10);
          if (!*e && (num > 0) && (num <= MAX_SUPPORT_DEVICES))
            devices[num - 1].setting.alarm_disable = atoi(q);
        }
        //
        // default config
        //
        else if (strcmp(p, DEFAULT_PULSE_TIME    ) == 0)
          defaults.pulse_time = atoi(q);
        else if (strcmp(p, DEFAULT_PULSE_INTERVAL) == 0)
          defaults.pulse_interval = atoi(q);
        else if (strcmp(p, DEFAULT_OPEN_INIT     ) == 0)
          defaults.open_init = atoi(q);
        else if (strcmp(p, DEFAULT_OPEN_TIME     ) == 0)
          defaults.open_time = atoi(q);
        else if (strcmp(p, DEFAULT_RESUME_TIME   ) == 0)
          defaults.resume_time = atoi(q);
        else if (strcmp(p, DEFAULT_AUTODW_TIME   ) == 0)
          defaults.autodw_time = atoi(q);
        else if (strcmp(p, DEFAULT_ALARM_DISABLE ) == 0)
          defaults.alarm_disable = atoi(q);
      }
    }
    fclose(fp);
  }
  else
    syslog(LOG_ERR, "[ERROR] open(\"%s\"): %s", buf, strerror(errno));
  for (i = 0; i < MAX_SUPPORT_DEVICES; ++i)
  {
    device_t& dev = devices[i];
    if (dev.setting.pulse_time     == (uint32_t)-1)
      dev.setting.pulse_time     = defaults.pulse_time;
    if (dev.setting.pulse_interval == (uint32_t)-1)
      dev.setting.pulse_interval = defaults.pulse_interval;
    if (dev.setting.open_init      == (uint32_t)-1)
      dev.setting.open_init      = defaults.open_init;
    if (dev.setting.open_time      == (uint32_t)-1)
      dev.setting.open_time      = defaults.open_time;
    if (dev.setting.resume_time    == (uint32_t)-1)
      dev.setting.resume_time    = defaults.resume_time;
    if (dev.setting.autodw_time    == (uint32_t)-1)
      dev.setting.autodw_time    = defaults.autodw_time;
    if (dev.setting.alarm_disable  == (uint32_t)-1)
      dev.setting.alarm_disable  = defaults.alarm_disable;
  }
}

