/*
 shutter.c - Garage Shutter Controller Manager

 Copyright (c) 2022 Sasapea's Lab. All right reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General
 Public License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <pthread.h>
#include <semaphore.h>
#include "mosquitto.h"
#include "config.h"
#include "ftdisio.h"
#include "bridge.h"

#define SERIAL_SPEED      1000000 // 1Mbps
#define CONNECT_ACTIVE    6
#define POLLING_INTERVAL  1000    // ms
#define IO_INTERVAL       100     // ms
#define IO_RETRY_COUNT    5

#define STOP  0
#define DOWN  1
#define UP    2
#define ALARM 3

#define lengthof(a) (sizeof(a) / sizeof(a[0]))

static const char TOPIC_CONTROL[] = "/home/garage/shutter/control/";
static const char TOPIC_POLLING[] = "/home/garage/shutter/polling/";
static const char TOPIC_QUERY[]   = "/home/garage/shutter/query/";
static const char TOPIC_STATE[]   = "/home/garage/shutter/state/";

typedef struct
{
  macaddr_t mac;
  int enable;
  int connect;
  int percent;
  int status;
} shutter_t;

shutter_t shutters[MAX_SUPPORT_DEVICES];

char            *prog_path;
char            *prog_name;
int              exitcode;
bool             debug;
pthread_t        thread_id;
FTDISIO          serial;
pthread_mutex_t  iolock;
sem_t            iowait;
int              result;
Bridge::packet_t rx_packet;
Bridge::packet_t tx_packet;

long long get_millis(struct timespec *ts)
{
  return (ts->tv_sec * 1000) + (ts->tv_nsec / 1000000);
}

void set_millis(struct timespec *ts, long long millis)
{
  ts->tv_sec  = (millis / 1000);
  ts->tv_nsec = (millis % 1000) * 1000000;
}

void add_millis(struct timespec *ts, long long millis)
{
  ts->tv_sec += (millis / 1000);
  if ((ts->tv_nsec += (millis % 1000) * 1000000) >= 1000000000)
  {
    ts->tv_nsec %= 1000000000;
    ts->tv_sec++;
  }
}

void delay(long millis)
{
  struct timespec ts;
  set_millis(&ts, millis);
  nanosleep(&ts, NULL);
}

void dump_packet(const Bridge::packet_t& packet, const char *prefix)
{
  if (debug)
  {
    const unsigned char *p = (const unsigned char *)&packet;
    pthread_mutex_lock(&iolock);
    if (prefix)
      printf("%s", prefix);
    for (int i = 0; i <= packet.length; ++i)
      printf("%02X ", *p++);
    printf("\n");
    pthread_mutex_unlock(&iolock);
  }
}

void init(void)
{
  static const macaddr_t MACZERO = {};
  load_config(prog_path);
  for (int i = 0; i < lengthof(shutters); ++i)
  {
    device_t&  device  = devices[i];
    shutter_t& shutter = shutters[i];
    memset(&shutter, 0, sizeof(shutter));
    if (memcmp(&device.mac, &MACZERO, sizeof(device.mac)) != 0)
    {
      memcpy(&shutter.mac, &device.mac, sizeof(shutter.mac));
      shutter.enable = 1;
    }
  }
}

char *get_topic(const char *topic, int no)
{
  static char buf[256];
  if (no < 0)
    snprintf(buf, sizeof(buf), "%s+", topic);
  else
    snprintf(buf, sizeof(buf), "%s%d", topic, no);
  return buf;
}

bool device_open(const char *name, long speed)
{
  if (serial.open(name, speed) < 0)
  {
    syslog(LOG_ERR, "[ERROR] open(\"%s\"): %s", name, strerror(errno));
    return false;
  }
  sem_init(&iowait, 0, 0);
  pthread_mutex_init(&iolock, NULL);
  return true;
}

void device_close(void)
{
  pthread_mutex_destroy(&iolock);
  sem_destroy(&iowait);
  serial.close();
}

int device_read(Bridge::packet_t& packet)
{
  ssize_t n, cnt;
  for (cnt = packet.length = 0; cnt <= packet.length; cnt += n)
  {
    if ((n = serial.read((uint8_t *)&packet + cnt, packet.length - cnt + sizeof(packet.length))) == 0)
      break;
  }
  if (cnt == packet.length + sizeof(packet.length))
  {
    //
    dump_packet(packet, "->");
    //
    if (Bridge::csum((uint8_t *)&packet, cnt) == 0)
    {
      --packet.length; // remove csum
      return 1;
    }
  }
  return n < 0 ? -1 : 0;
}

bool device_write(Bridge::packet_t& packet, bool wait)
{
  struct timespec timeout;
  ssize_t rv;

  for (int i = 0; i < IO_RETRY_COUNT; ++i)
  {
    // calculate csum
    //
    dump_packet(packet, "<-");
    //
    result = -1;
    if ((rv = serial.write(&packet, packet.length + sizeof(packet.length))) == -1)
    {
      syslog(LOG_ERR, "[ERROR] write(): %s", strerror(errno));
      break;
    }
    if (rv != packet.length + sizeof(packet.length))
    {
      syslog(LOG_ERR, "[ERROR] write(): %d != %d", rv, packet.length + sizeof(packet.length));
      break;
    }
    if (!wait)
    {
      result = 0;
      break;
    }
    clock_gettime(CLOCK_REALTIME, &timeout);
    add_millis(&timeout, IO_INTERVAL);
    if (sem_timedwait(&iowait, &timeout) == 0)
      break;
    // re-transmit
    if (!(packet.type & Bridge::TYPE_RETRANSMIT))
    {
      packet.type   = Bridge::TYPE_RETRANSMIT;
      packet.length = offsetof(__typeof__(packet), data);
      *((uint8_t *)&packet + packet.length) = Bridge::csum((uint8_t *)&packet, packet.length);
    }
  }
  return result != -1;
}

bool device_response(Bridge::packet_t& packet, struct mosquitto *mosq)
{
  char buf[16];

  command_t& response = *(command_t *)packet.data;
  if (packet.length >= offsetof(__typeof__(packet), data) + offsetof(__typeof__(response), args))
  {
    int no = ((uint16_t)packet.addrH << 8) | packet.addrL;
    if (no < lengthof(shutters))
    {
      shutter_t& shutter = shutters[no++];
      if (shutter.enable)
      {
        switch (response.command)
        {
          default:
            if (result == -1)
            {
              result = response.args.status;
              sem_post(&iowait);
            }
            shutter.connect = CONNECT_ACTIVE;
            break;
          case BRIDGE_CONNECT:
            if (response.args.status == 0)
            {
              shutter.percent = -1;
              shutter.status  = STOP;
              shutter.connect = CONNECT_ACTIVE;
              mosquitto_publish(mosq, NULL, get_topic(TOPIC_CONTROL, no),
                (int)snprintf(buf, sizeof(buf), "%s", "CONNECT"), buf, 0, false);
            }
            break;
          case BRIDGE_POLLING:
            if (response.args.status & 0x80)
            {
              response.args.status &= ~0x080;
              if (shutter.status != ALARM)
              {
                shutter.status = ALARM;
                syslog(LOG_INFO, "[INFO] shutter[%d] ALARM ON", no);
              }
            }
            else
            {
              if (shutter.status == ALARM)
              {
                shutter.status = STOP;
                syslog(LOG_INFO, "[INFO] shutter[%d] ALARM OFF", no);
              }
              if ((shutter.percent >= 0) && (response.args.status != shutter.percent))
              {
                if (response.args.status > shutter.percent)
                {
                  if (shutter.status != DOWN)
                  {
                    shutter.status = DOWN;
                    syslog(LOG_INFO, "[INFO] shutter[%d] CLOSING", no);
                  }
                }
                else
                {
                  if (shutter.status != UP)
                  {
                    shutter.status = UP;
                    syslog(LOG_INFO, "[INFO] shutter[%d] OPENING", no);
                  }
                }
                mosquitto_publish(mosq, NULL, get_topic(TOPIC_STATE, no),
                  (int)snprintf(buf, sizeof(buf), "%d", response.args.status), buf, 0, false);
              }
              else if ((shutter.percent < 0) || (shutter.status != STOP))
              {
                shutter.status = STOP;
                if (response.args.status == 100)
                  syslog(LOG_INFO, "[INFO] shutter[%d] CLOSED", no);
                else if (response.args.status == 0)
                  syslog(LOG_INFO, "[INFO] shutter[%d] OPENED", no);
                else
                  syslog(LOG_INFO, "[INFO] shutter[%d] STOPPED", no);
                mosquitto_publish(mosq, NULL, get_topic(TOPIC_STATE, no),
                  (int)snprintf(buf, sizeof(buf), "%d", response.args.status), buf, 0, false);
              }
            }
            shutter.percent = response.args.status;
            shutter.connect = CONNECT_ACTIVE;
            break;
        }
      }
    }
  }
  return true;
}

bool device_control(unsigned char cmd, int no, void *data, int len, bool wait)
{
  Bridge::packet_t& packet = tx_packet;
  command_t& request = *(command_t *)packet.data;
  if (cmd == BRIDGE_CONNECT)
  {
    packet.length   = sizeof(request.args.address);
    packet.type     = 0;
    packet.addrH    = 0xFF; // broadcast
    packet.addrL    = 0xFF; //
    request.command = cmd;
    memcpy(&request.args.address.mac, &shutters[no].mac, sizeof(request.args.address.mac));
    request.args.address.addrH = no >> 8;
    request.args.address.addrL = no >> 0;
  }
  else
  {
    packet.length   = len;
    packet.type     = 0;
    packet.addrH    = no >> 8;
    packet.addrL    = no >> 0;
    request.command = cmd;
    if (len)
      memcpy(request.args.data, data, len);
  }
  packet.length += offsetof(__typeof__(packet), data) + offsetof(__typeof__(request), args);
  *((uint8_t *)&packet + packet.length) = Bridge::csum((uint8_t *)&packet, packet.length);
  return device_write(packet, wait);
}

void mqtt_control(int no, unsigned char cmd, void *data, int len)
{
  for (int i = 0; i < lengthof(shutters); ++i)
  {
    if (((no == i + 1) || (no == 0)) && shutters[i].enable && (shutters[i].connect || (cmd == DEVICE_RESET)))
    {
      if (!device_control(cmd, i, data, len, true))
        syslog(LOG_ERR, "[ERROR] shutter[%d] control failed", i + 1);
    }
  }
}

void hton(void *buf, uint32_t val)
{
  ((uint8_t *)buf)[0] = val >> 24;
  ((uint8_t *)buf)[1] = val >> 16;
  ((uint8_t *)buf)[2] = val >>  8;
  ((uint8_t *)buf)[3] = val >>  0;
}

void mqtt_setting(int no, bool connect)
{
  for (int i = 0; i < lengthof(shutters); ++i)
  {
    if (((no == i + 1) || (no == 0)) && shutters[i].enable)
    {
      shutters[i].percent = -1;
      if (connect)
      {
      	setting_t  setting = {};
        device_t&  device  = devices[i];
        hton(&setting.pulse_time    , device.setting.pulse_time    );
        hton(&setting.pulse_interval, device.setting.pulse_interval);
        hton(&setting.open_init     , device.setting.open_init     );
        hton(&setting.open_time     , device.setting.open_time     );
        hton(&setting.resume_time   , device.setting.resume_time   );
        hton(&setting.autodw_time   , device.setting.autodw_time   );
        hton(&setting.alarm_disable , device.setting.alarm_disable );
      	if (!device_control(SHUTTER_SETTING, i, &setting, sizeof(setting), true))
          syslog(LOG_ERR, "[ERROR] shutter[%d] setting failed", i + 1);
      }
    }
  }
}

void mqtt_polling(int no)
{
  for (int i = 0; i < lengthof(shutters); ++i)
  {
    if (((no == i + 1) || (no == 0)) && shutters[i].enable)
    {
      int cmd = shutters[i].connect ? BRIDGE_POLLING : BRIDGE_CONNECT;
      if (!device_control(cmd, i, NULL, 0, false))
        syslog(LOG_ERR, "[ERROR] shutter[%d] polling failed", i + 1);
      delay(IO_INTERVAL);
    }
  }
}

void mqtt_alarm(int no, bool alarm)
{
  char buf[] = { (char)(no - 1) };
  for (int i = 0; i < lengthof(shutters); ++i)
  {
    if (((no == i + 1) || (no == 0)) && shutters[i].connect)
    {
      if (!device_control(alarm ? SHUTTER_PAUSE : SHUTTER_RESUME, i, buf, sizeof(buf), true))
        syslog(LOG_ERR, "[ERROR] shutter[%d] alarm notify failed", i + 1);
    }
  }
}

void mqtt_alarm_control(int no, bool enable)
{
  char buf[] = {enable};
  for (int i = 0; i < lengthof(shutters); ++i)
  {
    if (((no == i + 1) || (no == 0)) && shutters[i].connect)
    {
      if (!device_control(SHUTTER_ALARM, i, buf, sizeof(buf), true))
        syslog(LOG_ERR, "[ERROR] shutter[%d] alarm control failed", i + 1);
    }
  }
}

void mqtt_query(struct mosquitto *mosq, int no)
{
  char buf[16];
  for (int i = 0; i < lengthof(shutters); ++i)
  {
    if (((no == i + 1) || (no == 0)) && (shutters[i].percent != -1))
    {
      mosquitto_publish(mosq, NULL, get_topic(TOPIC_STATE, i + 1),
        (int)snprintf(buf, sizeof(buf), "%d", shutters[i].percent), buf, 0, false);
    }
  }
}

int mqtt_topic_is(const char *prefix, char *topic)
{
  int len = strlen(prefix);
  return strncmp(topic, prefix, len) == 0 ? atoi(topic + len) : -1;
}

int mqtt_command_is(const char *name, void *payload, int payloadlen)
{
  return (strlen(name) == payloadlen) && (strncmp(name, (char *)payload, payloadlen) == 0);
}

void mqtt_message_callback(struct mosquitto *mosq, void *userdata, const struct mosquitto_message *message)
{
  int no;
  if ((no = mqtt_topic_is(TOPIC_QUERY, message->topic)) >= 0)
    mqtt_query(mosq, no);
  else if ((no = mqtt_topic_is(TOPIC_POLLING, message->topic)) >= 0)
    mqtt_polling(no);
  else if ((no = mqtt_topic_is(TOPIC_CONTROL, message->topic)) >= 0)
  {
    syslog(LOG_INFO, "[INFO] shutter[%d] %.*s", no, message->payloadlen, (char *)message->payload);
    if (mqtt_command_is("CONNECT", message->payload, message->payloadlen))
      mqtt_setting(no, true);
    else if (mqtt_command_is("DISCONNECT", message->payload, message->payloadlen))
      mqtt_setting(no, false);
    else if (mqtt_command_is("STOP", message->payload, message->payloadlen))
      mqtt_control(no, SHUTTER_STOP, NULL, 0);
    else if (mqtt_command_is("DOWN", message->payload, message->payloadlen))
      mqtt_control(no, SHUTTER_DOWN, NULL, 0);
    else if (mqtt_command_is("UP", message->payload, message->payloadlen))
      mqtt_control(no, SHUTTER_UP, NULL, 0);
    else if (mqtt_command_is("AUTO_UP", message->payload, message->payloadlen))
      mqtt_control(no, SHUTTER_AUTO_UP, NULL, 0);
    else if (mqtt_command_is("PAUSE", message->payload, message->payloadlen))
      mqtt_control(no, SHUTTER_PAUSE, NULL, 0);
    else if (mqtt_command_is("RESUME", message->payload, message->payloadlen))
      mqtt_control(no, SHUTTER_RESUME, NULL, 0);
    else if (mqtt_command_is("ALARM_ON", message->payload, message->payloadlen))
      mqtt_alarm(no, true);
    else if (mqtt_command_is("ALARM_OFF", message->payload, message->payloadlen))
      mqtt_alarm(no, false);
    else if (mqtt_command_is("ALARM_ENABLE", message->payload, message->payloadlen))
      mqtt_alarm_control(no, true);
    else if (mqtt_command_is("ALARM_DISABLE", message->payload, message->payloadlen))
      mqtt_alarm_control(no, false);
    else if (mqtt_command_is("RESET", message->payload, message->payloadlen))
      mqtt_control(no, DEVICE_RESET, NULL, 0);
  }
}

void mqtt_connect_callback(struct mosquitto *mosq, void *userdata, int result)
{
  if (result == 0)
  {
    syslog(LOG_INFO, "[INFO] connected to mosquitto server");
    //
    // Setup Subscribe
    //
    mosquitto_subscribe(mosq, NULL, get_topic(TOPIC_CONTROL, -1), 0);
    mosquitto_subscribe(mosq, NULL, get_topic(TOPIC_POLLING, -1), 0);
    mosquitto_subscribe(mosq, NULL, get_topic(TOPIC_QUERY  , -1), 0);
  }
}

void mqtt_log_callback(struct mosquitto *mosq, void *userdata, int level, const char *str)
{
  switch (level)
  {
#if 0
    case MOSQ_LOG_DEBUG:
      syslog(LOG_DEBUG, "[DEBUG] mqtt_log_callback(): %s", str);
      break;
    case MOSQ_LOG_INFO:
      syslog(LOG_INFO, "[INFO] mqtt_log_callback(): %s", str);
      break;
#endif
    case MOSQ_LOG_NOTICE:
      syslog(LOG_NOTICE, "[NOTICE] mqtt_log_callback(): %s", str);
      break;
    case MOSQ_LOG_WARNING:
      syslog(LOG_WARNING, "[WARNING] mqtt_log_callback(): %s", str);
      break;
    case MOSQ_LOG_ERR: {
      syslog(LOG_ERR, "[ERROR] mqtt_log_callback(): %s", str);
      break;
    }
  }
}

void *polling(void *arg)
{
  struct mosquitto *mosq = (struct mosquitto *)arg;
  char buf[32];
  while (exitcode == 0)
  {
    delay(POLLING_INTERVAL);
    pthread_testcancel();
    for (int i = 0; i < lengthof(shutters); ++i)
    {
      shutter_t& shutter = shutters[i];
      if (shutter.connect)
      {
        if (--shutter.connect == 0)
          mosquitto_publish(mosq, NULL, get_topic(TOPIC_CONTROL, i + 1),
            (int)snprintf(buf, sizeof(buf), "%s", "DISCONNECT"), buf, 0, false);
      }
    }
    mosquitto_publish(mosq, NULL, get_topic(TOPIC_POLLING, 0), 0, NULL, 0, false);
  }
  return NULL;
}

int polling_start(struct mosquitto *mosq)
{
  return pthread_create(&thread_id, NULL, polling, mosq);
}

int polling_stop()
{
  pthread_cancel(thread_id);
  return pthread_join(thread_id, NULL);
}

void sigint_handler(int signum)
{
  exitcode = 1;
}

void usage()
{
  printf("Garage Shutter Controller Manager, Version 1.0\n");
  printf("Copyright(c) 2022 Sasapea's Lab. All right reserved.\n");
  printf("\n");
  printf("Usage: %s [options]\n", prog_name + 1);
  printf("\n");
  printf("  options:\n");
  printf("    -device    ... ftdi_sio device name (ttyUSB0)\n");
  printf("    -host      ... mqtt server name     (localhost)\n");
  printf("    -port      ... mqtt server port     (1883)\n");
  printf("    -keepalive ... mqtt keepalive       (60)\n");
  printf("    -shutter   ... shutter number       (*0/1/2/3)\n");
  printf("    -command   ... shutter command      (STOP/DOWN/UP/AUTO_UP/PAUSE/RESUME/RESET)\n");
  printf("    -recovery  ... sem_unlink\n");
  printf("    -help      ... this help\n\n");
}

int main(int argc, char **argv)
{
  const char *name = "ttyUSB0"; // ftdi_sio
  const char *host = "localhost";
  int port         = 1883;
  int keepalive    = 60;
  bool help        = false;
  bool recovery    = false;
  int  shutter     = 0;
  char *command    = NULL;
  struct mosquitto *mosq;
  struct sigaction nact, oact;
  sem_t *sem;
  int i;

  exitcode = 1;
  //
  // save this program path
  //
  prog_path = argv[0];
  prog_name = strrchr(prog_path, '/');
  //
  // load conf file
  //
  init();
  //
  // open syslog
  //
  openlog(prog_name + 1, LOG_CONS | LOG_PID | LOG_PERROR, LOG_USER);
  //
  // setup signal handler
  //
  sigemptyset(&nact.sa_mask);
  nact.sa_handler = sigint_handler;
  nact.sa_flags   = SA_RESTART;
  sigaction(SIGINT , &nact, &oact);
  sigaction(SIGTERM, &nact, &oact);
  //
  // Check Command Line Argument's
  //
  for (i = 1; i < argc; )
  {
    char *p = argv[i++];
    if (*p == '-')
    {
      ++p;
      if (strcmp(p, "debug") == 0)
        debug = true;
      else if (i < argc)
      {
        if (strcmp(p, "device") == 0)
          name = argv[i++];
        else if (strcmp(p, "host") == 0)
          host = argv[i++];
        else if (strcmp(p, "port") == 0)
          port = atoi(argv[i++]);
        else if (strcmp(p, "keepalive") == 0)
          keepalive = atoi(argv[i++]);
        else if (strcmp(p, "shutter") == 0)
          shutter = atoi(argv[i++]);
        else if (strcmp(p, "command") == 0)
          command = argv[i++];
        else
          help = true;
      }
      else if (strcmp(p, "recovery") == 0)
        recovery = true;
      else
        help = true;
    }
    else
      help = true;
  }
  //
  // Help
  //
  if (help)
    usage();
  else if (recovery)
  {
    //
    // Unlink Semaphote
    //
    exitcode = sem_unlink(prog_name) < 0;
  }
  else
  {
    //
    // Init MQTT Client Lib
    //
    mosquitto_lib_init();
    //
    // Publish Control Command
    //
    if (command)
    {
      //
      // New MQTT Cleint
      //
      if ((mosq = mosquitto_new(NULL, true, NULL)) == NULL)
        syslog(LOG_ERR, "[ERROR] mosquitto_new()");
      else
      {
        //
        // Connect MQTT Server
        //
        if (mosquitto_connect(mosq, host, port, keepalive) != MOSQ_ERR_SUCCESS)
          syslog(LOG_ERR, "[ERROR] mosquitto_connect(\"%s\", %d)", host, port);
        else
        {
          //
          // Publish Control Command
          //
          if (mosquitto_publish(mosq, NULL, get_topic(TOPIC_CONTROL, shutter),
              strlen(command), command, 0, false) == MOSQ_ERR_SUCCESS)
            exitcode = 0;
          //
          // Disconnect MQTT Server
          //
          mosquitto_disconnect(mosq);
        }
        //
        // Destory MQTT Client
        //
        mosquitto_destroy(mosq);
      }
    }
    //
    // Create Semaphote
    //
    else if ((sem = sem_open(prog_name, O_CREAT | O_EXCL, S_IRWXU, 0)) != SEM_FAILED)
    {
      //
      // Open Device
      //
      if (device_open(name, SERIAL_SPEED))
      {
        //
        // New MQTT Cleint
        //
        if ((mosq = mosquitto_new(prog_name, false, 0)) == NULL)
          syslog(LOG_ERR, "[ERROR] mosquitto_new()");
        else
        {
          //
          // Set MQTT Client Callback
          //
          mosquitto_connect_callback_set(mosq, mqtt_connect_callback);
          mosquitto_message_callback_set(mosq, mqtt_message_callback);
          mosquitto_log_callback_set(mosq, mqtt_log_callback);
          //
          // Connect MQTT Server
          //
          if (mosquitto_connect_async(mosq, host, port, keepalive) != MOSQ_ERR_SUCCESS)
            syslog(LOG_ERR, "[ERROR] mosquitto_connect_async(\"%s\", %d)", host, port);
          else
          {
            //
            // Start Polling Thread
            //
            exitcode = 0;
            if (polling_start(mosq) != 0)
              syslog(LOG_ERR, "[ERROR] polling_start(): %s", strerror(errno));
            else
            {
              //
              // Start MQTT Client
              //
              if (mosquitto_loop_start(mosq) != MOSQ_ERR_SUCCESS)
              {
                exitcode = 1;
                syslog(LOG_ERR, "[ERROR] mosquitto_loop_start()");
              }
              else
              {
                //
                // Read from Device
                //
                while (exitcode == 0)
                {
                  if ((i = device_read(rx_packet)) > 0)
                    device_response(rx_packet, mosq);
                  else if (i < 0)
                    exitcode = 1;
                }
                //
                // Stop MQTT Client
                //
                mosquitto_loop_stop(mosq, 1);
              }
              //
              // Stop Polling Thread
              //
              polling_stop();
            }
            //
            // Disconnect MQTT Server
            //
            mosquitto_disconnect(mosq);
          }
          //
          // Destory MQTT Client
          //
          mosquitto_destroy(mosq);
        }
        //
        // Close Device
        //
        device_close();
      }
      //
      // Close Semaphote
      //
      sem_close(sem);
      //
      // Unlink Semaphote
      //
      sem_unlink(prog_name);
    }
    else
    {
      syslog(LOG_INFO, "[INFO] process is already running");
    }
    //
    // Cleanup MQTT Client Lib
    //
    mosquitto_lib_cleanup();
  }
  //
  // print process exit code
  //
  syslog(LOG_DEBUG, "[DEBUG] terminated: %d", exitcode);
  //
  // close syslog
  //
  closelog();
  return exitcode;
}
